-- (C) 2015, Dominik Martin Drzic

-- Misc functions ------------------------------------------------------------------------------------------------------
local function SetToBounds (val, min, max)
	if min > max then
		min, max = max, min
	end

	if      val < min then return min
	elseif  val > max then return max
	else                   return val
	end
end

local function CheckBounds (val, min, max)
	return (val >= min and val <= max)
end

local function AreOverlapped (start1, end1, start2, end2)
	if start1 > end1 then start1, end1 = end1, start1 end
	if start2 > end2 then start2, end2 = end2, start2 end
	return (start1 <= end2 and start2 <= end1)
end

-- Save/restore muted MIDI events --------------------------------------------------------------------------------------
local function SaveMutedNotesToTable (table, take, itemStartPPQ, itemEndPPQ, sourceLenPPQ, looped)
	local id        = #table
	local noteCount = ({reaper.MIDI_CountEvts(take, 0, 0, 0)})[2]

	for i = 0, noteCount-1 do
		local _, _, muted, startPos, endPos, chan, pitch, vel = reaper.MIDI_GetNote(take, i)
		if muted then
			while (startPos <= itemEndPPQ) do
				id = id + 1
				table[id] = {}
				table[id].type     = "Note"
				table[id].startPos = reaper.MIDI_GetProjTimeFromPPQPos(take, SetToBounds(startPos, itemStartPPQ, itemEndPPQ))
				table[id].endPos   = reaper.MIDI_GetProjTimeFromPPQPos(take, SetToBounds(endPos,   itemStartPPQ, itemEndPPQ))
				table[id].chan     = chan
				table[id].pitch    = pitch
				table[id].vel      = vel
				if looped then
					startPos, endPos = startPos + sourceLenPPQ, endPos + sourceLenPPQ
				else
					break
				end
			end
		end
	end
end

local function SaveMutedCCToTable (table, take, itemStartPPQ, itemEndPPQ, sourceLenPPQ, looped)
	local id      = #table
	local ccCount = ({reaper.MIDI_CountEvts(take, 0, 0, 0)})[3]

	for i = 0, ccCount-1 do
		local _, _, muted, startPos, chanmsg, chan, msg2, msg3 = reaper.MIDI_GetCC(take, i)
		if muted then
			while (startPos <= itemEndPPQ) do
				id = id + 1
				table[id] = {}
				table[id].type     = "CC"
				table[id].startPos = reaper.MIDI_GetProjTimeFromPPQPos(take, SetToBounds(startPos, itemStartPPQ, itemEndPPQ))
				table[id].chanmsg  = chanmsg
				table[id].chan     = chan
				table[id].msg2     = msg2
				table[id].msg3     = msg3
				if looped then
					startPos, endPos = startPos + sourceLenPPQ, endPos + sourceLenPPQ
				else
					break
				end
			end
		end
	end
end

local function SaveMutedSysTextToTable (table, take, itemStartPPQ, itemEndPPQ, sourceLenPPQ, looped)
	local id           = #table
	local sysTextCount = ({reaper.MIDI_CountEvts(take, 0, 0, 0)})[4]

	for i = 0, sysTextCount-1 do
		local _, _, muted, startPos, sysType, string = reaper.MIDI_GetTextSysexEvt(take, i)
		if muted then
			while (startPos <= itemEndPPQ) do
				id = id + 1
				table[id] = {}
				table[id].type     = "TextSys"
				table[id].startPos = reaper.MIDI_GetProjTimeFromPPQPos(take, SetToBounds(startPos, itemStartPPQ, itemEndPPQ))
				table[id].sysType  = sysType
				table[id].string   = string
				if looped then
					startPos, endPos = startPos + sourceLenPPQ, endPos + sourceLenPPQ
				else
					break
				end
			end
		end
	end
end

local function SaveMutedEventsToTable (table, take)
	if reaper.BR_IsTakeMidi(take, 0) then
		local item  = reaper.GetMediaItemTake_Item(take)
		local itemStartPPQ = reaper.MIDI_GetPPQPosFromProjTime(take, reaper.GetMediaItemInfo_Value(item, "D_POSITION"))
		local itemEndPPQ   = reaper.MIDI_GetPPQPosFromProjTime(take, reaper.GetMediaItemInfo_Value(item, "D_POSITION") + reaper.GetMediaItemInfo_Value(item, "D_LENGTH"))
		local sourceLenPPQ = reaper.BR_GetMidiSourceLenPPQ(take)
		local looped       = reaper.GetMediaItemInfo_Value(item, "B_LOOPSRC") == 1

		local track = reaper.GetMediaItemTake_Track(take)
		if (table[track] == nil) then table[track] = {} end

		-- Muted events are saved regardless of item mix behavior in case of overlapping items
		SaveMutedNotesToTable(table[track], take, itemStartPPQ, itemEndPPQ, sourceLenPPQ, looped)
		SaveMutedCCToTable(table[track], take, itemStartPPQ, itemEndPPQ, sourceLenPPQ, looped)
		SaveMutedSysTextToTable(table[track], take, itemStartPPQ, itemEndPPQ, sourceLenPPQ, looped)
	end
end

local function SaveMutedEventsInSelectedItems (table)
	for i = 0, reaper.CountSelectedMediaItems(0)-1 do
		local item = reaper.GetSelectedMediaItem(0, i)

		if reaper.GetMediaItemInfo_Value(item, "B_ALLTAKESPLAY") == 1 then
			for j = 0, reaper.CountTakes(item)-1 do
				SaveMutedEventsToTable(table, reaper.GetTake(item, j))
			end
		else
			SaveMutedEventsToTable(table, reaper.GetActiveTake(item))
		end
	end
end

local function RestoreMutedMidiEventsToSelectedItems (table)
	for i = 0, reaper.CountTracks(0)-1 do
		local track = reaper.GetTrack(0, i)
		if table[track] ~= nill and #table[track] > 0 then

			-- Get all selected items in a track first (so we don't have to iterate through table multiple times for each item)
			local items = {}
			local itemsSize = 0
			for j = 0, reaper.CountTrackMediaItems(track)-1 do
				local item = reaper.GetTrackMediaItem(track, j)
				local take = reaper.GetActiveTake(item)
				if reaper.GetMediaItemInfo_Value(item, "B_UISEL") == 1 and reaper.BR_IsTakeMidi(take, 0) then
					items[itemsSize+1] = {}
					items[itemsSize+1].activeTake   = take
					items[itemsSize+1].startTime    = reaper.GetMediaItemInfo_Value(item, "D_POSITION")
					items[itemsSize+1].endTime      = reaper.GetMediaItemInfo_Value(item, "D_POSITION") + reaper.GetMediaItemInfo_Value(item, "D_LENGTH")
					items[itemsSize+1].needsSorting = false
					itemsSize = itemsSize + 1
				end
			end

			-- Insert muted events in all selected items in track
			local events   = table[track]
			for id in ipairs(events) do
				for index, _ in ipairs(items) do

					-- Check if event fits into current item (in case of MIDI note, if part of the note fits, bound it to item end position)
					local validForInsertion = false
					local startPos = events[id].startPos
					local endPos   = 0
					local take     = items[index].activeTake
					if reaper.BR_IsTakeMidi(take, 0) and startPos <= items[index].endTime then
						local itemStartTime = items[index].startTime
						local itemEndTime   = items[index].endTime
						if events[id].type == "Note" then
							startPos = SetToBounds(startPos,          itemStartTime, itemEndTime)
							endPos   = SetToBounds(events[id].endPos, itemStartTime, itemEndTime)
							validForInsertion = (endPos - startPos) > 0
						else
							validForInsertion = CheckBounds(startPos, itemStartTime, itemEndTime)
						end
					end

					-- Insert event
					if validForInsertion then
						startPos = reaper.MIDI_GetPPQPosFromProjTime(take, startPos)
						if events[id].type == "Note" then
							reaper.MIDI_InsertNote(take, false, true, startPos, reaper.MIDI_GetPPQPosFromProjTime(take, endPos), events[id].chan, events[id].pitch, events[id].vel, true)
						elseif events[id].type == "CC" then
							reaper.MIDI_InsertCC(take, false, true, startPos, events[id].chanmsg, events[id].chan, events[id].msg2, events[id].msg3)
						elseif events[id].type == "TextSys" then
							reaper.MIDI_InsertTextSysexEvt(take, false, true, startPos, events[id].sysType, events[id].string)
						end
						items[index].needsSorting = true
					end
				end
			end

			-- Make sure events are sorted after inserting
			for index in ipairs(items) do
				if items[index].needsSorting then
					reaper.MIDI_Sort(items[index].activeTake)
				end
			end
		end
	end
end

-- Save/restore item selection -----------------------------------------------------------------------------------------
local function SaveSelectedItemsToTable (table)
	for i = 0, reaper.CountSelectedMediaItems(0)-1 do
		table[i+1] = reaper.GetSelectedMediaItem(0, i)
	end
end

local function RestoreSelectedItemsFromTable (table)
	for _, item in ipairs(table) do
		reaper.SetMediaItemInfo_Value(item, "B_UISEL", 1)
	end
end

local function UnselectAllItems ()
	while (reaper.CountSelectedMediaItems(0) > 0) do
		reaper.SetMediaItemSelected(reaper.GetSelectedMediaItem(0, 0), false)
	end
end

local function SetItemSelected (item, unselectOthers)
	if unselectOthers then UnselectAllItems() end
	reaper.SetMediaItemSelected(item, true)
end

-- Save/restore items name,color and notes -----------------------------------------------------------------------------
local function SaveSelectedItemsData (table, saveNames, saveColors, saveSnapOffset, saveItemNotes)
	if saveNames or saveColors or saveItemNotes then
		for i = 0, reaper.CountTracks(0)-1 do

			local track = reaper.GetTrack(0, i)

			if (table[track] == nil) then table[track] = {} end
			local id = #table[track]

			for j = 0, reaper.CountTrackMediaItems(track)-1 do
				local item = reaper.GetTrackMediaItem(track, j)
				if reaper.GetMediaItemInfo_Value(item, "B_UISEL") == 1 then
					id = id + 1
					table[track][id] = {}
					local take  = reaper.GetActiveTake(item)
					local track = reaper.GetMediaItemTrack(item)

					table[track][id].startTime = reaper.GetMediaItemInfo_Value(item, "D_POSITION")
					table[track][id].endTime   = table[track][id].startTime + reaper.GetMediaItemInfo_Value(item, "D_LENGTH")

					 -- REAPER always glues to either MIDI or audio (so video item will end up as audio item)
					table[track][id].type = "MIDI"
					if not reaper.BR_IsTakeMidi(take, 0) then table[track][id].type = "AUDIO" end

					if saveNames then
						table[track][id].itemName = ({reaper.GetSetMediaItemTakeInfo_String(take, "P_NAME", "", false)})[2]
					end

					if saveColors then
						table[track][id].itemColor = reaper.GetMediaItemInfo_Value(item, "I_CUSTOMCOLOR")
						table[track][id].takeColor = reaper.GetMediaItemTakeInfo_Value(take, "I_CUSTOMCOLOR")
					end

					if saveSnapOffset then
						table[track][id].snapOffset = reaper.GetMediaItemInfo_Value(item, "D_SNAPOFFSET");
					end

					if saveItemNotes then
						table[track][id].itemNotes = reaper.ULT_GetMediaItemNote(item)
						if table[track][id].itemNotes == "" then table[track][id].itemNotes = nil end
						local imageValid, itemImage, itemImageFlags = reaper.BR_GetMediaItemImageResource(item, 0)
						if imageValid then
							table[track][id].itemImage      = itemImage;
							table[track][id].itemImageFlags = itemImageFlags;
						end
					end
				end
			end
		end
	end
end

local function RestoreItemDataToSelectedItems (table, restoreNames, restoreColors, restoreSnapOffset, restoreItemNotes)
	if restoreNames or restoreColors or restoreItemNotes then
		for i = 0, reaper.CountTracks(0)-1 do
			local track = reaper.GetTrack(0, i)
			if table[track] ~= nill and #table[track] > 0 then

				for j = 0, reaper.CountTrackMediaItems(track)-1 do
					local item = reaper.GetTrackMediaItem(track, j)
					local take = reaper.GetActiveTake(item)
					if reaper.GetMediaItemInfo_Value(item, "B_UISEL") == 1 then

						local startTime = reaper.GetMediaItemInfo_Value(item, "D_POSITION")
						local endTime   = startTime + reaper.GetMediaItemInfo_Value(item, "D_LENGTH")

						-- REAPER always glues to either MIDI or audio (so video item will end up as audio item)
						local currentType = "MIDI"
						if not reaper.BR_IsTakeMidi(take, 0) then currentType = "AUDIO" end

						-- Find id of saved item data that is best suited to be restored to current item (gluing can lower count of items
						-- so always prioritize first saved...except with item notes - we put item notes of multiple items into one item)
						local currentId = 0
						for index, object in ipairs(table[track]) do
							if object and object.type == currentType and AreOverlapped(startTime, endTime, object.startTime, object.endTime) then
								currentId = index
								break
							end
						end

						if table[track][currentId] then
							if restoreNames and table[track][currentId].itemName then
								reaper.GetSetMediaItemTakeInfo_String(take, "P_NAME", table[track][currentId].itemName, true)
							end

							if restoreColors then
								if table[track][currentId].itemColor then
									reaper.SetMediaItemInfo_Value(item, "I_CUSTOMCOLOR", table[track][currentId].itemColor)
								end
								if table[track][currentId].takeColor then
									reaper.SetMediaItemTakeInfo_Value(take, "I_CUSTOMCOLOR", table[track][currentId].takeColor)
								end
							end

							if restoreSnapOffset and table[track][currentId].snapOffset then
								reaper.SetMediaItemInfo_Value(item, "D_SNAPOFFSET", table[track][currentId].snapOffset)
							end

							if restoreItemNotes then
								local newItemNote    = nil
								local itemImage      = nil
								local itemImageFlags = nil
								for h=1, #table[track] do
									local object = table[track][h]
									if object and (object.itemNotes or object.itemImage) and object.type == currentType and AreOverlapped(startTime, endTime, object.startTime, object.endTime) then
										if newItemNote == nil then
											newItemNote = object.itemNotes
										else
											newItemNote = newItemNote .. "\r\n" .. object.itemNotes
										end

										if itemImage == nil then
											itemImage      = object.itemImage
											itemImageFlags = object.itemImageFlags
										end
									end
								end
								if newItemNote ~= nill then reaper.ULT_SetMediaItemNote(item, newItemNote) end
								if itemImage ~= nill and itemImageFlags ~= nill then reaper.BR_SetMediaItemImageResource(item, itemImage, itemImageFlags) end
							end
						end
					end
				end
			end
		end
	end
end

-- Glue functionality --------------------------------------------------------------------------------------------------
local function GetUndoStringAndCmd (processItemsSeparately, ignoreTimeSel, includeFades, preserveMutedMidiEvents, preserveColor, preserveNames, preserveSnapOffset, preserveItemNotes)
	local undoString = nil
	local cmd        = 0

	if         ignoreTimeSel and     includeFades then cmd = 40257 undoString = ", ignoring time selection, including leading fade-in and trailing fade-out"
	elseif     ignoreTimeSel and not includeFades then cmd = 40362 undoString = ", ignoring time selection"
	elseif not ignoreTimeSel and     includeFades then cmd = 40606 undoString = ", including leading fade-in and trailing fade-out"
	else                                               cmd = 41588 undoString = "" end

	if   processItemsSeparately then undoString = "Glue items separately" .. undoString
	else                             undoString = "Glue items"            .. undoString end

	local mutedMidiString = nil
	if preserveMutedMidiEvents then mutedMidiString = "muted MIDI events" end

	local itemDataStr     = nil
	if preserveColor or preserveNames or preserveSnapOffset or preserveItemNotes then
		local strings = {}
		if preserveColor      then strings[#strings + 1] = "color" end
		if preserveNames      then strings[#strings + 1] = "name"  end
		if preserveSnapOffset then strings[#strings + 1] = "snap offset" end
		if preserveItemNotes  then strings[#strings + 1] = "notes" end

		if #strings > 0 then
			for index, string in ipairs(strings) do
				if      index == 1        then itemDataStr = "item "     .. string
				else if index == #strings then itemDataStr = itemDataStr .. " and " .. string
				else                           itemDataStr = itemDataStr .. ", "    .. string end
				end
			end
		end
	end

	if         mutedMidiString and     itemDataStr then undoString = undoString .. " (preserve " .. mutedMidiString .. " and " .. itemDataStr .. ")"
	elseif     mutedMidiString and not itemDataStr then undoString = undoString .. " (preserve " .. mutedMidiString .. ")"
	elseif not mutedMidiString and     itemDataStr then undoString = undoString .. " (preserve " .. itemDataStr     .. ")" end
	return undoString, cmd
end

local function PerformGlue (cmd, preserveMutedMidiEvents, preserveColor, preserveNames, preserveSnapOffset, preserveItemNotes)
	local savedMutedEvents = {}
	local savedItemsData   = {}
	if preserveMutedMidiEvents then SaveMutedEventsInSelectedItems(savedMutedEvents) end
	SaveSelectedItemsData(savedItemsData, preserveNames, preserveColor, preserveSnapOffset, preserveItemNotes)

	reaper.Main_OnCommand(cmd, 0)
	if preserveMutedMidiEvents then RestoreMutedMidiEventsToSelectedItems(savedMutedEvents) end
	RestoreItemDataToSelectedItems(savedItemsData, preserveNames, preserveColor, preserveSnapOffset, preserveItemNotes)
end

local function GlueItems (doUndo, processItemsSeparately, ignoreTimeSel, includeFades, preserveMutedMidiEvents, preserveColor, preserveNames, preserveSnapOffset, preserveItemNotes)
	local processed = false

	local gluableItemCount = reaper.CountSelectedMediaItems(0)
	local emptyItems = {}
	for i = 0, reaper.CountSelectedMediaItems(0)-1 do
		local item = reaper.GetSelectedMediaItem(0, i)
		local take = reaper.GetActiveTake(item)
		if take == nil then
			gluableItemCount = gluableItemCount - 1
			emptyItems[#emptyItems + 1] = item
		end
	end

	if gluableItemCount > 0 then

		-- Prepare stuff
		local undoString, cmd = GetUndoStringAndCmd(processItemsSeparately, ignoreTimeSel, includeFades, preserveMutedMidiEvents, preserveColor, preserveNames, preserveSnapOffset, preserveItemNotes)
		if doUndo then reaper.Undo_BeginBlock() end
		reaper.PreventUIRefresh(1)

		-- Unselect all empty items (no point in gluing them)
		for _, item in ipairs(emptyItems) do
			reaper.SetMediaItemInfo_Value(item, "B_UISEL", 0)
		end

		-- Perform glue
		if processItemsSeparately then
			local savedItems = {}
			SaveSelectedItemsToTable(savedItems)
			for index, item in ipairs(savedItems) do
				SetItemSelected(item, true)
				PerformGlue(cmd, preserveMutedMidiEvents, preserveColor, preserveNames, preserveSnapOffset, preserveItemNotes)
				savedItems[index] = reaper.GetSelectedMediaItem(0, 0) -- just in case item pointer changes
			end
			UnselectAllItems()
			RestoreSelectedItemsFromTable(savedItems)
		else
			PerformGlue(cmd, preserveMutedMidiEvents, preserveColor, preserveNames, preserveSnapOffset, preserveItemNotes)
		end

		-- Restore empty items selection
		for _, item in ipairs(emptyItems) do
			reaper.SetMediaItemInfo_Value(item, "B_UISEL", 1)
		end

		-- Finish up
		reaper.PreventUIRefresh(-1)
		if doUndo then reaper.Undo_EndBlock(undoString, -1) end
		processed = true
	end
	return processed
end

-- Module export -------------------------------------------------------------------------------------------------------
local moduleExport = {}
moduleExport.GlueItems = GlueItems
return moduleExport